/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>

#ifdef __MINGW32__
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#endif

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "broccoli.h"
#include "bro_debug.h"
#include "bro_buf.h"
#include "bro_config.h"
#include "bro_hashtable.h"
#include "bro_types.h"
#include "bro_type.h"
#include "bro_val.h"
#include "bro_id.h"
#include "bro_event.h"
#include "bro_event_reg.h"
#include "bro_io.h"
#include "bro_util.h"
#include "bro_openssl.h"
#include "bro_record.h"
#include "bro_sem.h"
#include "bro_shm.h"
#ifdef BRO_PCAP_SUPPORT
#include "bro_packet.h"
#endif


/* Don't attempt to reconnect more than once every 5 seconds: */
#define BRO_RECONNECT_MAX_RATE   5

/* To make MinGW happy, we provide an initialization handler for the
 * DLL if we are building on Windows.
 */
#ifdef __MINGW32__
int main() { return 0; }
#endif

  
static int
conn_init_send_descriptor(BroConn *bc)
{
  uint32 *descriptor;
  
  D_ENTER;

  if (! (descriptor = calloc(4, sizeof(uint32))))
    D_RETURN_(FALSE);
  
  descriptor[0] = htonl((uint32) BRO_PROTOCOL_VERSION);
  descriptor[2] = htonl((uint32) BRO_DATA_FORMAT_VERSION);

  __bro_io_raw_queue(bc, BRO_MSG_VERSION,
		     (uchar*) descriptor, sizeof(uint32) * 4);
  
  free(descriptor);

  D_RETURN_(! bc->state->tx_dead);
}

static int
conn_init_complete(BroConn *bc)
{
  D_ENTER;
  
  /* If the user has requested handshake completion at this point, do so. */
  if ( (bc->conn_flags & BRO_CFLAG_COMPLETE_HANDSHAKE))
    {
      if (! bro_conn_await_handshake(bc, 10))
	D_RETURN_(FALSE);
    }
  
  D_RETURN_(TRUE);
}

static void
conn_init_io_handler(BroConn *bc)
{
  D_ENTER;

  bc->shared.io_pid = fork();
  
  if (bc->shared.io_pid < 0)
    {
      D(("Couldn't fork writer process: %s\n", strerror(errno)));
      bc->shared.io_pid = 0;
      D_RETURN;
    }
  
  if (bc->shared.io_pid != 0)
    D_RETURN;
  
  /* -- Child -- */
  
  if (! __bro_openssl_connect(bc))
    exit(-1);

  __bro_io_writer_loop(bc);
  
  D_RETURN;
}

static void
conn_kill_io_handler(BroConn *bc)
{
  int status;
  
  D_ENTER;

  __bro_sem_decr(bc->shared.io_sem_prod);
  __bro_sem_decr(bc->shared.io_sem);

  bc->state->io_msg = BRO_IOMSG_STOP;

  __bro_sem_incr(bc->shared.io_sem);
  __bro_sem_incr(bc->shared.io_sem_cons);	
  
  if (waitpid(bc->shared.io_pid, &status, 0) < 0)
    {
      D(("waitpid error waiting for %u: %s\n",
	 bc->shared.io_pid, strerror(errno)));
    }

  D_RETURN;
}


static int
conn_init_shared(BroConn *bc)
{
  D_ENTER;

  bc->shared.io_sem      = __bro_sem_new(bc);
  bc->shared.io_sem_prod = __bro_sem_new(bc);
  bc->shared.io_sem_cons = __bro_sem_new(bc);

  __bro_sem_attach(bc->shared.io_sem);
  __bro_sem_attach(bc->shared.io_sem_prod);
  __bro_sem_attach(bc->shared.io_sem_cons);
  
  if (! (bc->shared.rx_buf_shm = __bro_shm_new(0x4000)))
    goto error_exit;
  if (! (bc->shared.tx_buf_shm = __bro_shm_new(0x4000)))
    goto error_exit;
  if (! (bc->shared.state_shm  = __bro_shm_new(0x1000)))
    goto error_exit;
  
  if (! __bro_shm_attach(bc->shared.rx_buf_shm))
    goto error_exit;
  if (! __bro_shm_attach(bc->shared.tx_buf_shm))
    goto error_exit;
  if (! __bro_shm_attach(bc->shared.state_shm))
    goto error_exit;
  
  bc->rx_buf = __bro_buf_new_mem(__bro_shm_get_mem(bc->shared.rx_buf_shm),
				 __bro_shm_get_size(bc->shared.rx_buf_shm));
  bc->tx_buf = __bro_buf_new_mem(__bro_shm_get_mem(bc->shared.tx_buf_shm),
				 __bro_shm_get_size(bc->shared.tx_buf_shm));

  bc->state  = (BroConnState *) __bro_shm_get_mem(bc->shared.state_shm);
  bc->state->handshake_state = BRO_HANDSHAKE_NONE;
      
  /* Initially the tx buffer is empty, and the first producer can
   * fill it whenever he's ready. To allow this we lock the io_sem_cons
   * semaphore while io_sem_prod remains at 0.
   */
  __bro_sem_decr(bc->shared.io_sem_cons);
  
  /* With the connection and all synchronization mechanisms set up,
   * we now fork off a writer process if the user requested this
   * connection handle to be shareable across multiple processes/
   * threads.
   */
  conn_init_io_handler(bc);

  D_RETURN_(TRUE);

 error_exit:
  __bro_sem_detach(bc->shared.io_sem);
  __bro_sem_detach(bc->shared.io_sem_prod);
  __bro_sem_detach(bc->shared.io_sem_cons);
  
  __bro_shm_detach(bc->shared.rx_buf_shm);
  __bro_shm_detach(bc->shared.tx_buf_shm);
  __bro_shm_detach(bc->shared.state_shm);
  
  __bro_sem_free(bc->shared.io_sem);
  __bro_sem_free(bc->shared.io_sem_prod);
  __bro_sem_free(bc->shared.io_sem_cons);

  __bro_shm_free(bc->shared.rx_buf_shm);
  __bro_shm_free(bc->shared.tx_buf_shm);
  __bro_shm_free(bc->shared.state_shm);

  D_RETURN_(FALSE);
}

static void
conn_free_shared(BroConn *bc)
{
  /* The semaphores + shared memory are tricky to handle. Essentially,
   * we cannot predict where in the application bro_disconnect() might
   * be called, but it could certainly be in the process that created
   * the BroConn (and thus the semaphores + shared mem) or any process
   * forked from it. We therefore must prevent premature release of
   * the semaphores and shared mem while still allowing forked processes
   * to call bro_disconnect() to get rid of remaining memory.
   *
   * Our solution is to release the semaphores + shared mem only when
   * the process that originally created the BroConn is releasing it.
   * That means that this process *MUST*WAIT* until after all other processes
   * that share the connection have terminated before it calls bro_disconnect().
   */
  if (getpid() == bc->id_pid)
    {
      D(("Killing I/O handler\n"));
      conn_kill_io_handler(bc);
    }

  __bro_sem_detach(bc->shared.io_sem);
  __bro_sem_detach(bc->shared.io_sem_prod);
  __bro_sem_detach(bc->shared.io_sem_cons);

  __bro_shm_detach(bc->shared.rx_buf_shm);
  __bro_shm_detach(bc->shared.tx_buf_shm);
  __bro_shm_detach(bc->shared.state_shm);

      
  if (getpid() == bc->id_pid)
    {
      D(("Removing semaphores + shared mem for connection %p\n", bc));
      
      __bro_sem_free(bc->shared.io_sem);
      __bro_sem_free(bc->shared.io_sem_prod);
      __bro_sem_free(bc->shared.io_sem_cons);

      __bro_shm_free(bc->shared.rx_buf_shm);
      __bro_shm_free(bc->shared.tx_buf_shm);     
      __bro_shm_free(bc->shared.state_shm);
    }
}

static int
conn_init_excl(BroConn *bc)
{
  D_ENTER;

  if (! (bc->rx_buf = __bro_buf_new()))
    goto error_exit;
  if (! (bc->tx_buf = __bro_buf_new()))
    goto error_exit;

  if (! (bc->state = calloc(1, sizeof(BroConnState))))
    goto error_exit;
  
  bc->state->handshake_state = BRO_HANDSHAKE_NONE;

  if (! __bro_openssl_connect(bc))
    goto error_exit;

  D_RETURN_(TRUE);
  
 error_exit:
  __bro_buf_free(bc->rx_buf);
  __bro_buf_free(bc->tx_buf);
  
  D_RETURN_(FALSE);
}

static void
conn_free_excl(BroConn *bc)
{
  D_ENTER;

  __bro_openssl_shutdown(bc);

  if (bc->state)
    free(bc->state);

  __bro_buf_free(bc->rx_buf);
  __bro_buf_free(bc->tx_buf);

  D_RETURN;
}


BroConn *
bro_connect_str(const char *hostname, int flags)
{
  BroConn *bc;
  static int counter = 0;
  
  D_ENTER;

  if (! hostname || !*hostname)
    D_RETURN_(NULL);
  
  if (! (bc = (BroConn *) calloc(1, sizeof(BroConn))))
    D_RETURN_(NULL);
  
  D(("Connecting to host %s\n", hostname));

  bc->conn_flags = flags;
  bc->id_pid = getpid();
  bc->id_num = counter++;
  bc->peer = strdup(hostname);

  TAILQ_INIT(&bc->msg_queue);
  bc->msg_queue_len = 0;

  if (! (bc->ev_reg = __bro_event_reg_new()))
    goto error_exit;

  if (! (bc->io_cache = __bro_ht_new(__bro_ht_int_hash,
				     __bro_ht_int_cmp,
				     NULL,
				     (BroHTFreeFunc) __bro_sobject_release)))
    goto error_exit;
  
  if (! (bc->data = __bro_ht_new(__bro_ht_str_hash,
				 __bro_ht_str_cmp,
				 __bro_ht_mem_free,
				 NULL)))
    goto error_exit;
  
  if (! (bc->ev_mask = __bro_ht_new(__bro_ht_str_hash,
				    __bro_ht_str_cmp,
				    __bro_ht_mem_free,
				    NULL)))
    goto error_exit;  

  if ( (flags & BRO_CFLAG_SHAREABLE))
    {
      if (! conn_init_shared(bc))
	goto error_exit;
    }
  else
    {
      if (! conn_init_excl(bc))
	goto error_exit;
    }
  
  if (! conn_init_send_descriptor(bc))
    goto error_exit;
  if (! conn_init_complete(bc))
    goto error_exit;
  
  D_RETURN_(bc);

 error_exit:
  __bro_event_reg_free(bc->ev_reg);
  __bro_ht_free(bc->ev_mask);
  __bro_ht_free(bc->io_cache);
  __bro_ht_free(bc->data);

  if (bc->state)
    free(bc->state);
  
  if (bc->peer)
    free(bc->peer);
  
  free(bc);
  D_RETURN_(NULL);
}


BroConn *
bro_connect(struct in_addr *ip_addr, uint16 port, int flags)
{
  BroConn *bc;
  char hostname[1024];

  D_ENTER;
  snprintf(hostname, 1024, "%s:%hu", inet_ntoa(*ip_addr), ntohs(port));
  bc = bro_connect_str(hostname, flags);
  D_RETURN_(bc);
}


int
bro_reconnect(BroConn *bc)
{
  BroMsg *msg, *msg_first, **msg_last;
  int old_flags, msg_queue_len;
  time_t current_time;

  D_ENTER;

  if (! bc)
    D_RETURN_(FALSE);

  if (bc->state->in_reconnect)
    D_RETURN_(FALSE);
  
  if ( (current_time = time(NULL)) > 0)
    {
      if (current_time - bc->state->last_reconnect < BRO_RECONNECT_MAX_RATE)
	{
	  D(("Less than %i seconds since last reconnect attempt, not reconnecting.\n",
	     BRO_RECONNECT_MAX_RATE));
	  D_RETURN_(FALSE);
	}

      bc->state->last_reconnect = current_time;
    }
  
  D(("Attempting reconnection...\n"));
  bc->state->in_reconnect = TRUE;
  
  /* NOTE: the sequencing in this function is quite tricky and very picky.
   * Don't move things around unneccesarily ...
   */

  old_flags = bc->conn_flags;
  bc->conn_flags &= ~BRO_CFLAG_COMPLETE_HANDSHAKE;
  bc->state->tx_dead = bc->state->rx_dead = FALSE;
  
  /* Okay first clean up the old stuff: */
  bc->state->handshake_state = BRO_HANDSHAKE_NONE;

  __bro_buf_reset(bc->rx_buf);
  __bro_buf_reset(bc->tx_buf);
  
  if (bc->bio)
    {
      BIO_free_all(bc->bio);
      bc->bio = NULL;
    }

  /* Attempt to establish new connection */
  if (! __bro_openssl_reconnect(bc))
    goto error_return;

  /* Only *after* we managed to connect, we clear the event mask for events
   * the peer expects, etc. If we do this earlier and fail to connect, then future
   * sent events never trigger a reconnect because they're never sent since
   * they don't seem to be requested.
   */

  /* Temporarily unhook messages from the message queue so the new messages
   * get sent right away. We hook the old ones back in below.
   */
  msg_first = bc->msg_queue.tqh_first;
  msg_last  = bc->msg_queue.tqh_last;
  msg_queue_len = bc->msg_queue_len;
  bc->msg_queue_len = 0;
  TAILQ_INIT(&bc->msg_queue);

  __bro_ht_free(bc->ev_mask);

  if (! (bc->ev_mask = __bro_ht_new(__bro_ht_str_hash,
				    __bro_ht_str_cmp,
				    __bro_ht_mem_free,
				    NULL)))
    goto reset_error_return;

  __bro_ht_free(bc->io_cache);
  
  if (! (bc->io_cache = __bro_ht_new(__bro_ht_int_hash,
				     __bro_ht_int_cmp,
				     NULL,
				     (BroHTFreeFunc) __bro_sobject_release)))
    goto reset_error_return;

  /* NOTE: thanks to OpenSSL fuckedupness, we do not know at this point
   * whether the connection is actually established or not in the unencrypted
   * case.
   */
  if (! conn_init_send_descriptor(bc))
    goto reset_error_return;

  if (! conn_init_complete(bc))
    goto reset_error_return;
    
  /* Request events as before */
  __bro_event_reg_request(bc);

  /* Now complete handshake */
  if (! bro_conn_await_handshake(bc, 10))
    goto reset_error_return;

  /* Hook old events back in */
  if (bc->msg_queue_len == 0)
    bc->msg_queue.tqh_first = msg_first;
  else
    {
      msg_first->msg_queue.tqe_prev = bc->msg_queue.tqh_last;
      *bc->msg_queue.tqh_last = msg_first;
    }

  bc->msg_queue.tqh_last = msg_last;
  bc->msg_queue_len += msg_queue_len;

  D(("Reconnect completed successfully.\n"));
  bc->conn_flags = old_flags;
  bc->state->in_reconnect = FALSE;
  
  D_RETURN_(TRUE);

 reset_error_return:

  /* If the reconnect went wrong somehow, nuke the queue and
   * place old queue contents back in.
   */
  while ( (msg = bc->msg_queue.tqh_first))
    {
      TAILQ_REMOVE(&bc->msg_queue, msg, msg_queue);
      __bro_io_msg_free(msg);
    }
  
  bc->msg_queue.tqh_first = msg_first;
  bc->msg_queue.tqh_last = msg_last;
  bc->msg_queue_len = msg_queue_len;
  
 error_return:
  bc->state->tx_dead = bc->state->rx_dead = TRUE;
  bc->conn_flags = old_flags;
  bc->state->in_reconnect = FALSE;

  D_RETURN_(FALSE);
}


int 
bro_disconnect(BroConn *bc)
{
  BroMsg *msg;

  if (!bc)
    return FALSE;

  if (! bc->state->tx_dead)
    {
      /* Try to flush command queue */
      __bro_io_msg_queue_flush(bc);
    }

  while ( (msg = bc->msg_queue.tqh_first))
    {
      TAILQ_REMOVE(&bc->msg_queue, msg, msg_queue);
      __bro_io_msg_free(msg);
    }

  __bro_ht_free(bc->ev_mask);
  __bro_event_reg_free(bc->ev_reg);
  __bro_ht_free(bc->io_cache);
  __bro_ht_free(bc->data);

  if ( (bc->conn_flags & BRO_CFLAG_SHAREABLE))
    conn_free_shared(bc);
  else
    conn_free_excl(bc);

  if (bc->peer)
    free(bc->peer);

  free(bc);
  
  return TRUE;
}


int
bro_conn_await_handshake(BroConn *bc, int max_sec)
{
  int i = 0, intervals = max_sec * 5 + 1;
  
  D_ENTER;
  
  if (! bc || bc->state->rx_dead || max_sec < 0)
    D_RETURN_(FALSE);

  if (! __bro_io_send_handshake_complete(bc))
    {
      D_RETURN_(FALSE);
    }
  
  if (bc->state->handshake_state == BRO_HANDSHAKE_COMPLETE)
    D_RETURN_(TRUE);
  
  /* If we get here we're in state BRO_HANDSHAKE_SENT, and
   * wait until we receive the handshake-complete message from
   * the peer.
   */

  while (i < intervals)
    {
      struct timeval timeout;
      timeout.tv_sec  = 0;
      timeout.tv_usec = 200000;
      
      if (select(0, NULL, NULL, NULL, &timeout) < 0)
	{
	  D(("select() caused '%s'\n", strerror(errno)));
	  D_RETURN_(FALSE);
	}
      
      __bro_io_process_input(bc);
      
      if (bc->state->handshake_state == BRO_HANDSHAKE_COMPLETE)
	D_RETURN_(TRUE);
      
      if (max_sec != 0)
	i++;
    }
  
  D_RETURN_(FALSE);
}


int
bro_conn_alive(const BroConn *bc)
{
  if (! bc)
    return FALSE;

  return (! bc->state->tx_dead && ! bc->state->rx_dead);
}


static int
conn_adopt_events_cb(char *ev_name, void *data, BroHT *dst)
{
  char *key;

  /* If the event is already reported, ignore. */
  if (__bro_ht_get(dst, ev_name))
    return TRUE;
  
  D(("Adopting event %s\n", ev_name));
  key = strdup(ev_name);
  __bro_ht_add(dst, key, key);
  
  return TRUE;
  data = NULL;
}

void
bro_conn_adopt_events(BroConn *src, BroConn *dst)
{
  D_ENTER;

  if (!src || !dst)
    D_RETURN;

  __bro_ht_foreach(src->ev_mask, (BroHTCallback) conn_adopt_events_cb, dst->ev_mask);

  D_RETURN;
}


int
bro_conn_get_fd(BroConn *bc)
{
  int fd;

  if (! bc || bc->state->tx_dead || bc->state->rx_dead)
    return -1;
  
#ifdef __MINGW32__
  return 0;
#else
  BIO_get_fd(bc->bio, &fd);
  return fd;
#endif
}


int
bro_conn_process_input(BroConn *bc)
{
  if (! bc || bc->state->rx_dead)
    return FALSE;

  return __bro_io_process_input(bc);
}


int
bro_event_queue_length(BroConn *bc)
{
  if (! bc)
    return 0;

  return bc->msg_queue_len;
}


int
bro_event_queue_length_max(BroConn *bc)
{
  return BRO_MSG_QUEUELEN_MAX;
  bc = NULL;
}


int            
bro_event_queue_flush(BroConn *bc)
{
  return __bro_io_msg_queue_flush(bc);
}


int
bro_event_send(BroConn *bc, BroEvent *ev)
{
  int result;

  D_ENTER;

  if (! bc || !ev)
    {
      D(("Input error\n"));
      D_RETURN_(FALSE);
    }

  /* D(("Peer reports %i different event type(s).\n", __bro_ht_get_size(bc->ev_mask))); */
  
  result = __bro_io_event_queue(bc, ev);
  D_RETURN_(result);
}


void
bro_conn_data_set(BroConn *bc, const char *key, void *val)
{
  if (!bc || !key || !*key)
    return;

  __bro_ht_add(bc->data, strdup(key), val);
}


void *
bro_conn_data_get(BroConn *bc, const char *key)
{
  if (!bc || !key || !*key)
    return NULL;

  return __bro_ht_get(bc->data, (void *) key);
}


void *
bro_conn_data_del(BroConn *bc, const char *key)
{
  if (!bc || !key || !*key)
    return NULL;
  
  return __bro_ht_del(bc->data, (void *) key);
}


/* ----------------------------- Bro Events -------------------------- */

BroEvent *
bro_event_new(const char *event_name)
{
  BroString name;
  BroEvent *result;

  bro_string_set(&name, event_name);
  result = __bro_event_new(&name);
  bro_string_cleanup(&name);

  return result;
}


void
bro_event_free(BroEvent *be)
{
  __bro_event_free(be);
}


int
bro_event_add_val(BroEvent *be, int type, const void *val)
{
  BroVal *v;

  D_ENTER;

  if (! be || ! val || type < 0 || type >= BRO_TYPE_MAX)
    {
      D(("Invalid input: (%p, %i, %p)\n", be, type, val));
      D_RETURN_(FALSE);
    }
  
  if (! (v = __bro_val_new_of_type(type)))
    D_RETURN_(FALSE);

  if (! __bro_val_assign(v, val))
    {
      __bro_sobject_release((BroSObject *) v);
      D_RETURN_(FALSE);
    }

  __bro_event_add_val(be, v);

  D_RETURN_(TRUE);
}


int
bro_event_set_val(BroEvent *be, int val_num,
		  int type, const void *val)
{
  BroVal *v;
  int result;

  D_ENTER;

  if (! be || ! val || type < 0 || type >= BRO_TYPE_MAX)
    {
      D(("Invalid input: (%p, %i, %p)\n", be, type, val));
      D_RETURN_(FALSE);
    }

  if (! (v = __bro_val_new_of_type(type)))
    D_RETURN_(FALSE);

  if (! __bro_val_assign(v, val))
    {
      __bro_sobject_release((BroSObject *) v);
      D_RETURN_(FALSE);
    }
  
  result = __bro_event_set_val(be, val_num, v);
  D_RETURN_(result);
}


void
bro_event_registry_add(BroConn *bc,
		       const char *event_name,
		       BroEventFunc func)
{
  __bro_event_reg_add(bc, event_name, func);
}


void
bro_event_registry_remove(BroConn *bc, const char *event_name)
{
  __bro_event_reg_remove(bc, event_name);
}


void
bro_event_registry_request(BroConn *bc)
{
  __bro_event_reg_request(bc);
}



/* ------------------------ Dynamic-size Buffers --------------------- */

BroBuf *
bro_buf_new(void)
{
  BroBuf *buf;

  if (! (buf = calloc(1, sizeof(BroBuf))))
    return NULL;

  __bro_buf_init(buf);
  return buf;
}


void
bro_buf_free(BroBuf *buf)
{
  if (!buf)
    return;

  __bro_buf_cleanup(buf);
  free(buf);
}


int
bro_buf_append(BroBuf *buf, void *data, int data_len)
{
  return __bro_buf_append(buf, data, data_len);
}


void
bro_buf_consume(BroBuf *buf)
{
  __bro_buf_consume(buf);
}


uchar *
bro_buf_get(BroBuf *buf)
{
  return __bro_buf_get(buf);
}


uchar *
bro_buf_get_end(BroBuf *buf)
{
  return __bro_buf_get_end(buf);
}


uint
bro_buf_get_size(BroBuf *buf)
{
  return __bro_buf_get_size(buf);
}


uint
bro_buf_get_used_size(BroBuf *buf)
{
  return __bro_buf_get_used_size(buf);
}


uchar *
bro_buf_ptr_get(BroBuf *buf)
{
  return __bro_buf_ptr_get(buf);
}


uint32
bro_buf_ptr_tell(BroBuf *buf)
{
  return __bro_buf_ptr_tell(buf);
}


int
bro_buf_ptr_seek(BroBuf *buf, int offset, int whence)
{
  return __bro_buf_ptr_seek(buf, offset, whence);
}


int
bro_buf_ptr_check(BroBuf *buf, int size)
{
  return __bro_buf_ptr_check(buf, size);
}


int
bro_buf_ptr_read(BroBuf *buf, void *data, int size)
{
  return __bro_buf_ptr_read(buf, data, size);
}


int
bro_buf_ptr_write(BroBuf *buf, void *data, int size)
{
  return __bro_buf_ptr_write(buf, data, size);
}



/* ------------------------ Configuration Access --------------------- */

void
bro_conf_set_domain(const char *domain)
{
  __bro_conf_set_domain(domain);
}


int
bro_conf_get_int(const char *val_name, int *val)
{
  if (! val_name || ! val)
    return FALSE;

  return __bro_conf_get_int(val_name, val);
}


int
bro_conf_get_dbl(const char *val_name, double *val)
{
  if (! val_name || ! val)
    return FALSE;

  return __bro_conf_get_dbl(val_name, val);
}


const char *
bro_conf_get_str(const char *val_name)
{
  if (! val_name)
    return FALSE;
  
  return __bro_conf_get_str(val_name);
}


/* -------------------------- Record Handling ------------------------ */

BroRecord *
bro_record_new(void)
{
  return __bro_record_new();
}


void
bro_record_free(BroRecord *rec)
{
  return __bro_record_free(rec);
}

 
int
bro_record_add_val(BroRecord *rec, const char *name,
		   int type, const void *val)
{
  BroVal *v;

  D_ENTER;

  if (! rec || !name || !*name)
    {
      D(("Input error: (%p, %s, %i, %p)\n", rec, name, type, val));
      D_RETURN_(FALSE);
    }
  
  if (! (v = __bro_val_new_of_type(type)))
    {
      D(("Could not get val of type %i\n", type));
      D_RETURN_(FALSE);
    }
  
  __bro_sobject_data_set((BroSObject *) v, "field", strdup(name));

  if (! __bro_val_assign(v, val))
    {
      D(("Could not assign value to the new val.\n"));
      __bro_sobject_release((BroSObject *) v);
      D_RETURN_(FALSE);
    }

  __bro_record_add_val(rec, v);
  D_RETURN_(TRUE);
}


int
bro_record_get_nth_val(BroRecord *rec, int num,
			 int type, void *result)
{
  BroVal *val;
  int type_found;

  if (type < 0 || type >= BRO_TYPE_MAX || ! result)
    return FALSE;

  if (! (val = __bro_record_get_nth_val(rec, num)))
    return FALSE;

  /* Now transform the val into a form expected in *result,
   * based on the type given in the val.
   */
  if (! __bro_val_get_data(val, &type_found, result))
    return FALSE;
  
  if (type_found != type)
    {
      D(("Type mismatch: expected type tag %i, found type tag %i\n", type, type_found));
      return FALSE;
    }
  
  return TRUE;
}


int
bro_record_get_named_val(BroRecord *rec, const char *name,
			   int type, void *result)
{
  BroVal *val;
  int type_found;
 
  if (type < 0 || type >= BRO_TYPE_MAX || ! result)
    return FALSE;
 
  if (! (val = __bro_record_get_named_val(rec, name)))
    return FALSE;

  /* Now transform the val into a form expected in *result,
   * based on the type given in the val.
   */
  if (! __bro_val_get_data(val, &type_found, result))
    return FALSE;
  
  if (type_found != type)
    {
      D(("Type mismatch: expected type tag %i, found type tag %i\n", type, type_found));
      return FALSE;
    }
  
  return TRUE;
}


int
bro_record_set_nth_val(BroRecord *rec, int num,
		       int type, const void *val)
{
  BroVal *v;
  char *name;

  D_ENTER;

  if (! rec || num < 0 || num >= rec->val_len ||
      type < 0 || type >= BRO_TYPE_MAX || ! val)
    {
      D(("Input error: (%p, %i, %i, %p)\n", rec, num, type, val));
      D_RETURN_(FALSE);
    }
  
  if (! (v = __bro_record_get_nth_val(rec, num)))
    D_RETURN_(FALSE);

  if (! (name = __bro_sobject_data_get((BroSObject *) v, "field")))
    D_RETURN_(FALSE);
  
  if (! (v = __bro_val_new_of_type(type)))
    {
      D(("Could not get val of type %i\n", type));
      D_RETURN_(FALSE);
    }
  
  __bro_sobject_data_set((BroSObject *) v, "field", strdup(name));

  if (! __bro_val_assign(v, val))
    {
      D(("Could not assign value to the new val.\n"));
      __bro_sobject_release((BroSObject *) v);
      D_RETURN_(FALSE);
    }

  __bro_record_set_nth_val(rec, num, v);
  D_RETURN_(TRUE);
}


int
bro_record_set_named_val(BroRecord *rec, const char *name,
			 int type, const void *val)
{
  BroVal *v;

  D_ENTER;

  if (! rec || ! name || !*name ||
      type < 0 || type >= BRO_TYPE_MAX || ! val)
    {
      D(("Input error: (%p, %s, %i, %p)\n", rec, name, type, val));
      D_RETURN_(FALSE);
    }
  
  if (! (v = __bro_val_new_of_type(type)))
    {
      D(("Could not get val of type %i\n", type));
      D_RETURN_(FALSE);
    }
  
  __bro_sobject_data_set((BroSObject *) v, "field", strdup(name));
  
  if (! __bro_val_assign(v, val))
    {
      D(("Could not assign value to the new val.\n"));
      __bro_sobject_release((BroSObject *) v);
      D_RETURN_(FALSE);
    }
  
  __bro_record_set_named_val(rec, name, v);
  D_RETURN_(TRUE);
}


/* ------------------------------ Strings ---------------------------- */

void
bro_string_init(BroString *bs)
{
  if (! bs)
    return;

  memset(bs, 0, sizeof(BroString));
}


int
bro_string_set(BroString *bs, const char *s)
{
  if (! bs || !s)
    return FALSE;
  
  return bro_string_set_data(bs, s, strlen(s));
}


int
bro_string_set_data(BroString *bs, const uchar *data, int data_len)
{
  uchar *data_copy;
  
  if (! bs || !data || data_len < 0)
    return FALSE;
  
  if (! (data_copy = malloc(data_len + 1)))
    return FALSE;
  
  memcpy(data_copy, data, data_len);
  data_copy[data_len] = '\0';
  
  bs->str_len = data_len;
  bs->str_val = data_copy;

  return TRUE;
}


BroString *
bro_string_copy(BroString *bs)
{
  BroString *result;

  if (! bs)
    return NULL;

  if (! (result = calloc(1, sizeof(BroString))))
    return NULL;

  result->str_len = bs->str_len;

  if (! (result->str_val = malloc(result->str_len + 1)))
    {
      free(result);
      return NULL;
    }

  memcpy(result->str_val, bs->str_val, result->str_len);
  result->str_val[result->str_len] = '\0';
  return result;
}


void
bro_string_cleanup(BroString *bs)
{
  if (! bs)
    return;

  if (bs->str_val)
    free(bs->str_val);

  memset(bs, 0, sizeof(BroString));
}


void
bro_string_free(BroString *bs)
{
  if (! bs)
    return;

  bro_string_cleanup(bs);
  free(bs);
}

/* ----------------------- Pcap Packet Handling ---------------------- */
#ifdef BRO_PCAP_SUPPORT

void
bro_conn_set_packet_ctxt(BroConn *bc, int link_type)
{
  if (! bc)
    return;

  bc->pcap_link_type = link_type;
}


void
bro_conn_get_packet_ctxt(BroConn *bc, int *link_type)
{
  if (! bc)
    return;

  if (link_type)
    *link_type = bc->pcap_link_type;
}


BroPacket *
bro_packet_new(const struct pcap_pkthdr *hdr, const u_char *data)
{
  BroPacket *packet;

  if (! hdr || ! data)
    return NULL;

  if (! (packet = calloc(1, sizeof(BroPacket))))
    return NULL;

  packet->pkt_pcap_hdr = *hdr;

  if (! (packet->pkt_data = malloc(hdr->caplen)))
    {
      free(packet);
      return NULL;
    }

  memcpy((u_char *) packet->pkt_data, data, hdr->caplen);
  return packet;
}


BroPacket *
bro_packet_clone(const BroPacket *src)
{
  BroPacket *dst;
  
  if (! (dst = calloc(1, sizeof(BroPacket))))
    return NULL;
  
  if (! __bro_packet_clone(dst, src))
    {
      bro_packet_free(dst);
      return NULL;
    }

  return dst;
}


void
bro_packet_free(BroPacket *packet)
{
  if (! packet)
    return;

  if (packet->pkt_data)
    free((u_char *) packet->pkt_data);

  free(packet);
}

int
bro_packet_send(BroConn *bc, BroPacket *packet)
{
  if (! bc || ! packet)
    {
      D(("Invalid input.\n"));
      return FALSE;
    }

  return __bro_io_packet_queue(bc, packet);
}

#endif

/* --------------------------- Miscellaneous ------------------------- */

double
bro_util_current_time(void)
{
  return __bro_util_get_time();
}


double 
bro_util_timeval_to_double(const struct timeval *tv)
{
  return __bro_util_timeval_to_double(tv);
}



