/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#ifdef __EMX__
#include <strings.h>
#endif 

#include <bro_types.h>
#include <bro_type.h>
#include <bro_val.h>
#include <bro_debug.h>
#include <bro_event.h>

BroEvent      *
__bro_event_new(BroString *name)
{
  BroEvent *ev;
  BroString *name_copy;

  if (!name)
    return NULL;

  if (! (ev = calloc(1, sizeof(BroEvent))))
    return NULL;

  TAILQ_INIT(& (ev->val_list));
  
  if (! (name_copy = bro_string_copy(name)))
    {
      free(ev);
      return NULL;
    }

  ev->name = *name_copy;

  return ev;
}


void           
__bro_event_free(BroEvent *ev)
{
  BroVal *val;

  if (!ev)
    return;

  bro_string_cleanup(&ev->name);

  while ( (val = ev->val_list.tqh_first))
    {
      TAILQ_REMOVE(&ev->val_list, val, val_list);
      __bro_sobject_release((BroSObject *) val);
    }
  
  free(ev);
}


BroEvent      *
__bro_event_copy(BroEvent *be)
{
  BroEvent *be_copy;
  BroVal *val, *val_copy;
  
  if (! be)
    return NULL;
  
  if (! (be_copy = __bro_event_new(&be->name)))
    return NULL;
  
  for (val = be->val_list.tqh_first; val; val = val->val_list.tqe_next)
    {
      if (! (val_copy = (BroVal *) __bro_sobject_copy((BroSObject *) val)))
	{
	  __bro_event_free(be_copy);
	  return NULL;
	}
      
      __bro_event_add_val(be_copy, val_copy);
    }
  
  return be_copy;
}


const char *
__bro_event_get_name(BroEvent *be)
{
  if (! be)
    return NULL;

  return be->name.str_val;
}


int
__bro_event_serialize(BroEvent *ev, BroConn *bc)
{
  BroVal *val;

  D_ENTER;

  /* Prepare the beginning of a serialized event call --
   * event identifier, event name plus argument description.
   */
  if (! __bro_buf_write_char(bc->tx_buf, 'e'))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_string(bc->tx_buf, &ev->name))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_double(bc->tx_buf, __bro_util_get_time()))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_int(bc->tx_buf, ev->val_len))
    D_RETURN_(FALSE);
    
  /* Now serialize each remaining parameter into the buffer: */
  for (val = ev->val_list.tqh_first; val; val = val->val_list.tqe_next)
    {
      if (! __bro_sobject_serialize((BroSObject *) val, bc))
	D_RETURN_(FALSE);
    }

  D_RETURN_(TRUE);
}


BroEvent *
__bro_event_unserialize(BroConn *bc)
{
  int i;
  BroString ev_name;
  double ev_time;
  uint32 ev_args;
  BroEvent *ev = NULL;
  BroVal *val = NULL;

  D_ENTER;
  
  if (! __bro_buf_read_string(bc->rx_buf, &ev_name))
    {
      D(("Couldn't read event name.\n"));
      D_RETURN_(NULL);
    }
  
  if (! __bro_buf_read_double(bc->rx_buf, &ev_time))
    {
      D(("Couldn't read event time.\n"));
      bro_string_cleanup(&ev_name);
      D_RETURN_(NULL);
    }
  
  if (! __bro_buf_read_int(bc->rx_buf, &ev_args))
    {
      D(("Couldn't read number of event arguments.\n"));
      bro_string_cleanup(&ev_name);
      D_RETURN_(NULL);
    }
  
  D(("Reading %i arguments for event %s\n", ev_args, ev_name.str_val));
  ev = __bro_event_new(&ev_name);
  bro_string_cleanup(&ev_name);
  
  for (i = 0; i < (int) ev_args; i++)
    {
      D(("Reading parameter %i\n", i+1));
      if (! (val = (BroVal *) __bro_sobject_unserialize(SER_IS_VAL, bc)))
	{
	  D(("Couldn't read parameter val %i.\n", i+1));
	  __bro_event_free(ev);
	  D_RETURN_(NULL);
	}
      
      __bro_event_add_val(ev, val);
    }

  D_RETURN_(ev);
}


void
__bro_event_add_val(BroEvent *ev, BroVal *v)
{
  D_ENTER;
  TAILQ_INSERT_TAIL(&ev->val_list, v, val_list);
  ev->val_len++;
  D_RETURN;
}


int
__bro_event_set_val(BroEvent *be, int val_num, BroVal *v)
{
  int i;
  BroVal *val;

  D_ENTER;

  if (val_num < 0 || val_num >= be->val_len)
    {
      D(("Invalid val index: given %i, having %i elements.\n", val_num, be->val_len));
      D_RETURN_(FALSE);
    }
  
  for (i = 0, val = be->val_list.tqh_first; val; i++, val = val->val_list.tqe_next)
    {
      if (i == val_num)
	{
	  /* First insert after the current element ... */
	  TAILQ_INSERT_AFTER(&be->val_list, val, v, val_list);
	  
	  /* ... and then remove the current one. */
	  TAILQ_REMOVE(&be->val_list, val, val_list);

	  /* Release the old value. */
	  __bro_sobject_release((BroSObject *) val);
	  
	  D_RETURN_(TRUE);
	}
    }
  
  D_RETURN_(FALSE);
}
