/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#ifdef __EMX__
#include <strings.h>
#endif 

#include <bro_types.h>
#include <bro_debug.h>
#include <bro_hashtable.h>

#define BRO_HT_NUM_SLOTS 127

#ifdef BRO_DEBUG
#define DEBUG_HTS
#endif

#ifdef DEBUG_HTS
#  define DEBUG_HT(x) do { printf("%s/%i: ", __FILE__, __LINE__); printf x ; } while (0);
#else
#  define DEBUG_HT(x)
#endif


typedef struct bro_hash_item
{
  TAILQ_ENTRY(bro_hash_item) hi_list;

  void         *it_key;
  void         *it_data;

} BroHTIt;

typedef TAILQ_HEAD(hi_list, bro_hash_item) BroHIList;

struct bro_ht
{
  BroHIList    *ht_slots;
  int           ht_numslots;
  int           ht_size;

  BroHTHashFunc ht_hash_func;
  BroHTCmpFunc  ht_cmp_func;
  BroHTFreeFunc ht_key_free_func;
  BroHTFreeFunc ht_val_free_func;
};


BroHT *
__bro_ht_new(BroHTHashFunc hash_func,
	     BroHTCmpFunc cmp_func,
	     BroHTFreeFunc key_free_func,
	     BroHTFreeFunc val_free_func)
{
  BroHT *ht;
  int i;

  /* It may be okay if we get no free_funcs ... */
  if (!hash_func || !cmp_func)
    return NULL;

  if (! (ht = calloc(1, sizeof(BroHT))))
    return NULL;

  ht->ht_numslots = BRO_HT_NUM_SLOTS;
  ht->ht_size = 0;
  ht->ht_hash_func = hash_func;
  ht->ht_cmp_func = cmp_func;
  ht->ht_key_free_func = key_free_func;
  ht->ht_val_free_func = val_free_func;
  
  if (! (ht->ht_slots = malloc(ht->ht_numslots * sizeof(BroHIList))))
    {
      free(ht);
      return NULL;
    }
  
  for (i = 0; i < ht->ht_numslots; i++)
    TAILQ_INIT(&ht->ht_slots[i]);
  
  return ht;
}


void
__bro_ht_free(BroHT *ht)
{
  BroHTIt *it;
  int i;

  if (!ht)
    return;

  for (i = 0; i < ht->ht_numslots; i++)
    {
      while ( (it = ht->ht_slots[i].tqh_first))
	{
	  TAILQ_REMOVE(&ht->ht_slots[i], it, hi_list);

	  if (ht->ht_key_free_func)
	    ht->ht_key_free_func(it->it_key);
	  if (ht->ht_val_free_func)
	    ht->ht_val_free_func(it->it_data);

	  free(it);
	}
    }
  
  free(ht->ht_slots);
  free(ht);
}


void
__bro_ht_add(BroHT *ht, void *key, void *data)
{
  uint32 slot;
  BroHTIt *it;

  if (!ht || !key)
    {
      D(("Input error: (%p, %p, %p)\n", ht, key, data));
      return;
    }
  
  if (! (it = calloc(1, sizeof(BroHTIt))))
    {
      D(("Out of memory.\n"));
      return;
    }
  
  it->it_key = key;
  it->it_data = data;
  
  slot = ht->ht_hash_func(key) % ht->ht_numslots;
  TAILQ_INSERT_TAIL(&ht->ht_slots[slot], it, hi_list);
  ht->ht_size++;
}


void *
__bro_ht_get(BroHT *ht, const void *key)
{
  BroHTIt *it;
  uint32 slot;
  
  if (!ht || !key)
    {
      D(("Input error: (%p, %p)\n", ht, key));
      return NULL;
    }
  
  slot = ht->ht_hash_func(key) % ht->ht_numslots;
    
  for (it = ht->ht_slots[slot].tqh_first; it; it = it->hi_list.tqe_next)
    {
      if (ht->ht_cmp_func(it->it_key, key))
	return it->it_data;
    }

  return NULL;
}


void     *
__bro_ht_del(BroHT *ht, void *key)
{
  void *result;
  BroHTIt *it;
  uint32 slot;

  if (!ht || !key)
    return NULL;

  slot = ht->ht_hash_func(key) % ht->ht_numslots;
  
  for (it = ht->ht_slots[slot].tqh_first; it; it = it->hi_list.tqe_next)
    {
      if (ht->ht_cmp_func(it->it_key, key))
	{
	  result = it->it_data;
	  TAILQ_REMOVE(&ht->ht_slots[slot], it, hi_list);
	  ht->ht_size--;
	  
	  /* Free the key if possible -- don't free the
	   * value, as we just return that.
	   */
	  if (ht->ht_key_free_func)
	    ht->ht_key_free_func(it->it_key);
	  
	  free(it);
	  
	  return result;
	}
    }
  
  return NULL;
}


int
__bro_ht_get_size(BroHT *ht)
{
  if (! ht)
    return -1;

  return ht->ht_size;
}


void
__bro_ht_foreach(BroHT *ht, BroHTCallback callback, void *user_data)
{
  BroHTIt *it;
  int i;

  if (!ht || !callback)
    return;
  
  for (i = 0; i < ht->ht_numslots; i++)
    {
      for (it = ht->ht_slots[i].tqh_first; it; it = it->hi_list.tqe_next)
	if (! callback(it->it_key, it->it_data, user_data))
	  return;
    }
}


uint32
__bro_ht_str_hash(const void *val)
{
  char *val_ptr;
  uint32 hash;
  
  if (!val)
    return 0;

  val_ptr = (char *) val;
  
  for (hash = 0; *val_ptr != '\0'; val_ptr++)
    hash = (64 * hash + *val_ptr);

  return hash;
}


int
__bro_ht_str_cmp(const void *val1, const void *val2)
{
  return (strcmp((char*) val1, (char*) val2) == 0);
}


void
__bro_ht_mem_free(void *data)
{
  if (data)
    free(data);
}


uint32
__bro_ht_int_hash(const void *val)
{
  return (uint32) val;
}


int
__bro_ht_int_cmp(const void *val1, const void *val2)
{
  return (val1 == val2);
}
