/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#ifdef __EMX__
#include <strings.h>
#endif 

#include <bro_types.h>
#include <bro_type.h>
#include <bro_io.h>
#include <bro_debug.h>
#include <bro_val.h>
#include <bro_record.h>


BroVal *
__bro_val_new(void)
{
  BroVal *val;

  D_ENTER;

  if (! (val = calloc(1, sizeof(BroVal))))
    D_RETURN_(NULL);

  __bro_val_init(val);

  D_RETURN_(val);
}


BroVal *
__bro_val_new_of_type(int type)
{
  BroVal *val;

  D_ENTER;

  switch (type)
    {
    case BRO_TYPE_BOOL:
    case BRO_TYPE_INT:
    case BRO_TYPE_COUNT:
    case BRO_TYPE_COUNTER:
    case BRO_TYPE_DOUBLE:
    case BRO_TYPE_TIME:
    case BRO_TYPE_INTERVAL:
    case BRO_TYPE_STRING:
    case BRO_TYPE_TIMER:
    case BRO_TYPE_PORT:
    case BRO_TYPE_IPADDR:
    case BRO_TYPE_NET:
    case BRO_TYPE_SUBNET:
      if (! (val = __bro_val_new()))
	D_RETURN_(NULL);
      break;
      
    case BRO_TYPE_RECORD:
      if (! (val = (BroVal *) __bro_record_val_new()))
	D_RETURN_(NULL);
      break;
      
    case BRO_TYPE_PATTERN:
    case BRO_TYPE_ENUM:
    case BRO_TYPE_ANY:
    case BRO_TYPE_TABLE:
    case BRO_TYPE_UNION:
    case BRO_TYPE_LIST:
    case BRO_TYPE_FUNC:
    case BRO_TYPE_FILE:
    case BRO_TYPE_VECTOR:
    case BRO_TYPE_ERROR:
    default:
      D(("Unsupported value type %i\n", type));
      D_RETURN_(NULL);
    }
  
  if(! (val->val_type = __bro_type_new_of_type(type)))
    {
      __bro_val_free(val);
      D_RETURN_(NULL);
    }
  
  D_RETURN_(val);
}


int
__bro_val_assign(BroVal *val, const void *data)
{
  D_ENTER;

  if (! val)
    {
      D(("Input error: (%p, %p)\n", val, data));
      D_RETURN_(FALSE);
    }
  
  if (! data)
    {
      if (val->val_type)
	{
	  __bro_sobject_release((BroSObject *) val->val_type);
	  val->val_type = NULL;
	}

      D(("Marked val %p as unassigned.\n", val));
      D_RETURN_(TRUE);
    }


  /* If we intend to assign data to the val, it must have a type. */
  if (! val->val_type)
    {
      D(("Cannot assign to val without a type.\n"));
      D_RETURN_(FALSE);
    }

  switch (val->val_type->tag)
    {
    case BRO_TYPE_BOOL:
      {
	int tmp = *((int *) data);
	val->val.char_val = (tmp != 0 ? 1 : 0);
      }
      break;

    case BRO_TYPE_INT:
    case BRO_TYPE_COUNT:
    case BRO_TYPE_COUNTER:
      val->val_int = *((int *) data);
      break;

    case BRO_TYPE_DOUBLE:
    case BRO_TYPE_TIME:
    case BRO_TYPE_INTERVAL:
      val->val_double = *((double *) data);
      break;

    case BRO_TYPE_STRING:
      {
	BroString *copy = bro_string_copy((BroString *) data);
	val->val_str = *copy;
      }
      break;
      
    case BRO_TYPE_PORT:
      {
	BroPort *tmp = (BroPort *) data;
	
	if (tmp->port_proto != IPPROTO_TCP &&
	    tmp->port_proto != IPPROTO_UDP &&
	    tmp->port_proto != IPPROTO_ICMP)
	  {
	    __bro_sobject_release((BroSObject *) data);
	    D_RETURN_(FALSE);
	  }
	
	val->val_port = *tmp;
      }
      break;
      
    case BRO_TYPE_IPADDR:
    case BRO_TYPE_NET:
      val->val_int = *((uint32 *) data);
      break;
      
    case BRO_TYPE_SUBNET:
      val->val_subnet = *((BroSubnet *) data);
      break;
      
    case BRO_TYPE_RECORD:
      {
	BroVal *tmp_val;
	BroRecordVal *rv = (BroRecordVal *) val;
	BroRecord *rec = (BroRecord *) data;
	
	if (rv->rec)
	  __bro_record_free(rv->rec);
	
	rv->rec = __bro_record_copy(rec);
	
	for (tmp_val = rec->val_list.tqh_first; tmp_val; tmp_val = tmp_val->val_list.tqe_next)
	  {
	    BroType *type;
	    char *field;

	    if (! tmp_val->val_type)
	      {
		D(("Cannot create record type component from val without type.\n"));
		D_RETURN_(FALSE);
	      }
	    
	    if (! (field = __bro_sobject_data_get((BroSObject *) tmp_val, "field")))
	      {
		D(("Val in record doesn't have field name associated with it.\n"));
		D_RETURN_(FALSE);
	      }

	    if (! (type = (BroType *) __bro_sobject_copy((BroSObject *) tmp_val->val_type)))
	      D_RETURN_(FALSE);

	    __bro_record_type_add_type((BroRecordType *) val->val_type, field, type);
	  }
      }
      break;

    case BRO_TYPE_PATTERN:
    case BRO_TYPE_ENUM:
    case BRO_TYPE_TIMER:
    case BRO_TYPE_ANY:
    case BRO_TYPE_TABLE:
    case BRO_TYPE_UNION:
    case BRO_TYPE_LIST:
    case BRO_TYPE_FUNC:
    case BRO_TYPE_FILE:
    case BRO_TYPE_VECTOR:
    case BRO_TYPE_ERROR:
      D(("Type %i currently unsupported.\n", val->val_type->tag));
      D_RETURN_(FALSE);
      
    default:
      D(("Unknown type identifier %i\n", val->val_type->tag));
      D_RETURN_(FALSE);
    }

  D_RETURN_(TRUE);
}


void
__bro_val_init(BroVal *val)
{
  BroSObject *sobj = (BroSObject *) val;

  D_ENTER;
  
  __bro_object_init((BroObject *) val);
  
  sobj->read  = (BroSObjectRead) __bro_val_read;
  sobj->write = (BroSObjectWrite) __bro_val_write;
  sobj->free  = (BroSObjectFree) __bro_val_free;
  sobj->clone = (BroSObjectClone) __bro_val_clone;

  /* Note: we don't know yet what type_id we'll be using since
   * that will depend on the type object hooked into this val.
   * We take care of that when we're serializing this val out
   * in that case.
   */
  sobj->type_id = SER_VAL;

  val->get_data = __bro_val_get;

  D_RETURN;
}


void
__bro_val_free(BroVal *val)
{
  D_ENTER;

  /* If there is no type in the val, then it's unassigned and
   * hence there won't be anything to clean up anyway.
   */
  if (val->val_type)
    {
      switch (val->val_type->tag)
	{
	case BRO_TYPE_STRING:
	  bro_string_cleanup(&val->val_str);
	  break;
	  
	default:
	  /* Nothing to do */
	  break;
	}
    }
  
  __bro_sobject_release((BroSObject *) val->val_type);
  __bro_object_free((BroObject *) val);

  D_RETURN;
}


int
__bro_val_get_type_num(const BroVal *val)
{
  if (! val)
    return 0;

  return val->val_type->tag;
}


int
__bro_val_get_data(BroVal *val, int *type, void **data)
{
  D_ENTER;

  if (! val || ! data)
    D_RETURN_(FALSE);

  if (! val->get_data)
    D_RETURN_(FALSE);

  if (type && val->val_type)
    *type = val->val_type->tag;
  
  *data = val->get_data(val);
  D_RETURN_(TRUE);
}


int
__bro_val_read(BroVal *val, BroConn *bc)
{
  char opt;
  int tmp;

  D_ENTER;

  if (! val || !bc)
    D_RETURN_(FALSE);

  if (! __bro_object_read((BroObject *) val, bc))
    D_RETURN_(FALSE);

  /* Read type */

  if (val->val_type)
    {
      __bro_sobject_release((BroSObject *) val->val_type);
      val->val_type = NULL;
    }

  if (! (val->val_type = (BroType *) __bro_sobject_unserialize(SER_IS_TYPE, bc)))
    D_RETURN_(FALSE);

  D(("Type in val has type tags %i/%i\n",
     val->val_type->tag, val->val_type->internal_tag));
     
  /* Read optional Attributes */
  
  if (val->val_attrs)
    {
      __bro_sobject_release((BroSObject *) val->val_attrs);
      val->val_attrs = NULL;
    }

  if (! __bro_buf_read_char(bc->rx_buf, &opt))
    D_RETURN_(FALSE);
  if (opt)
    {
      if (! (val->val_attrs = (BroAttrs *) __bro_sobject_unserialize(SER_ATTRIBUTES, bc)))
	D_RETURN_(FALSE);
    }
  
  switch (val->val_type->internal_tag)
    {
    case BRO_INTTYPE_INT:
    case BRO_INTTYPE_UNSIGNED:
      /* Hack for ports */
      if (val->val_type->tag == BRO_TYPE_PORT)
	{
	  if (! __bro_buf_read_int(bc->rx_buf, &tmp))
	    D_RETURN_(FALSE);
	  
	  if ( (tmp & 0x10000))
	    val->val_port.port_proto = IPPROTO_UDP;
	  else if ( (tmp & 0x20000))
	    val->val_port.port_proto = IPPROTO_ICMP;
	  else
	    val->val_port.port_proto = IPPROTO_TCP;
	    
	  val->val_port.port_num = (tmp & 0xFFFF);
	}
      else
	{
	  if (! __bro_buf_read_int(bc->rx_buf, &val->val_int))
	    D_RETURN_(FALSE);
	}
      break;
      
    case BRO_INTTYPE_DOUBLE:
      if (! __bro_buf_read_double(bc->rx_buf, &val->val_double))
	D_RETURN_(FALSE);
      break;
      
    case BRO_INTTYPE_STRING:
      if (! __bro_buf_read_string(bc->rx_buf, &val->val_str))
	D_RETURN_(FALSE);
      break;

    case BRO_INTTYPE_IPADDR:
      if (! __bro_buf_read_int(bc->rx_buf, &tmp))
	D_RETURN_(FALSE);
      
      if (tmp != 1)
	{
	  D(("We don't handle IPv6 addresses yet.\n"));
	  D_RETURN_(FALSE);
	}
      
      if (! __bro_buf_read_int(bc->rx_buf, &val->val_int))
	D_RETURN_(FALSE);

      val->val_int = ntohl(val->val_int);
      break;

    case BRO_INTTYPE_SUBNET:
      if (! __bro_buf_read_int(bc->rx_buf, &tmp))
	D_RETURN_(FALSE);
      
      if (tmp != 1)
	{
	  D(("We don't handle IPv6 addresses yet.\n"));
	  D_RETURN_(FALSE);
	}

      if (! __bro_buf_read_int(bc->rx_buf, &val->val_subnet.sn_net))
	D_RETURN_(FALSE);
      val->val_subnet.sn_net = ntohl(val->val_subnet.sn_net);

      if (! __bro_buf_read_int(bc->rx_buf, &val->val_subnet.sn_width))
	D_RETURN_(FALSE);
      break;

    case BRO_INTTYPE_OTHER:
      /* See Val.cc around 165 -- these are handled by derived classes.
       * We only make sure here it's not functions and not files.
       */
      if (val->val_type->tag != BRO_TYPE_FUNC &&
	  val->val_type->tag != BRO_TYPE_FILE)
	break;
      
      /* Otherwise fall through to warning. */

    default:
      D(("Unsupported internal type tag: %i\n", val->val_type->internal_tag));
      D_RETURN_(FALSE);
    }
  
  D_RETURN_(TRUE);
}


int
__bro_val_write(BroVal *val, BroConn *bc)
{
  BroType *type;
  BroSObject *obj;

  D_ENTER;
  
  if (! val || !bc)
    D_RETURN_(FALSE);

  /* We need to make sure that the BroSObject at the root has the
   * correct type_id (a SER_xxx value). This depends on the type object
   * so map the type tag of that object to a SER_xxx value:
   */
  if (! val->val_type)
    {
      D(("Val %p doesn't have a type.\n", val));
      D_RETURN_(FALSE);
    }

  type = (BroType *) val->val_type;
  obj  = (BroSObject *) val;
  
  switch (type->tag)
    {
    case BRO_TYPE_BOOL:
    case BRO_TYPE_INT:
    case BRO_TYPE_ENUM:
    case BRO_TYPE_COUNT:
    case BRO_TYPE_COUNTER:
    case BRO_TYPE_STRING:
    case BRO_TYPE_DOUBLE:
    case BRO_TYPE_TIME:
      obj->type_id = SER_VAL;
      break;
      
    case BRO_TYPE_PORT:
      obj->type_id = SER_PORT_VAL;
      break;
      
    case BRO_TYPE_INTERVAL:
      obj->type_id = SER_INTERVAL_VAL;
      break;
      
    case BRO_TYPE_IPADDR:
      obj->type_id = SER_ADDR_VAL;
      break;
      
    case BRO_TYPE_NET:
      obj->type_id = SER_NET_VAL;
      break;
      
    case BRO_TYPE_SUBNET:
      obj->type_id = SER_SUBNET_VAL;
      break;

    case BRO_TYPE_RECORD:
      obj->type_id = SER_RECORD_VAL;
      break;
      
    default:
      D(("Val %p's type unhandled: type tag is %i.\n", val, type->tag));
      D_RETURN_(FALSE);
    }

  if (! __bro_object_write((BroObject *) val, bc))
    D_RETURN_(FALSE);

  if (! __bro_sobject_serialize((BroSObject *) val->val_type, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_write_char(bc->tx_buf, val->val_attrs ? 1 : 0))
    D_RETURN_(FALSE);
  
  if (val->val_attrs && ! __bro_sobject_serialize((BroSObject *) val->val_attrs, bc))
    D_RETURN_(FALSE);
  
  switch (val->val_type->internal_tag)
    {
    case BRO_INTTYPE_INT:
    case BRO_INTTYPE_UNSIGNED:
      /* Hack for ports */
      if (val->val_type->tag == BRO_TYPE_PORT)
	{
	  int tmp = val->val_port.port_num;
	  
	  if (val->val_port.port_proto == IPPROTO_UDP)
	    tmp |= 0x10000;
	  else if (val->val_port.port_proto == IPPROTO_ICMP)
	    tmp |= 0x20000;

	  if (! __bro_buf_write_int(bc->tx_buf, tmp))
	    D_RETURN_(FALSE);
	}
      else
	{
	  if (! __bro_buf_write_int(bc->tx_buf, val->val_int))
	    D_RETURN_(FALSE);
	}
      break;
      
    case BRO_INTTYPE_DOUBLE:
      if (! __bro_buf_write_double(bc->tx_buf, val->val_double))
	D_RETURN_(FALSE);
      break;
      
    case BRO_INTTYPE_STRING:
      if (! __bro_buf_write_string(bc->tx_buf, &val->val_str))
	D_RETURN_(FALSE);
      break;

    case BRO_INTTYPE_IPADDR:
      if (! __bro_buf_write_int(bc->tx_buf, 1))
	D_RETURN_(FALSE);
      if (! __bro_buf_write_int(bc->tx_buf, htonl(val->val_int)))
	D_RETURN_(FALSE);
      break;

    case BRO_INTTYPE_SUBNET:
      if (! __bro_buf_write_int(bc->tx_buf, 1))
	D_RETURN_(FALSE);
      if (! __bro_buf_write_int(bc->tx_buf, htonl(val->val_subnet.sn_net)))
	D_RETURN_(FALSE);
      if (! __bro_buf_write_int(bc->tx_buf, val->val_subnet.sn_width))
	D_RETURN_(FALSE);
      break;

    case BRO_INTTYPE_OTHER:
      /* That's fine, will be handled in derived classes
       * like __bro_record_val_write().
       */
      break;
      
    default:
      D(("Unknown internal type tag: %i\n", val->val_type->internal_tag));
      D_RETURN_(FALSE);
    }  
  
  D_RETURN_(TRUE);
}

int
__bro_val_clone(BroVal *dst, BroVal *src)
{
  D_ENTER;
  
  if (! __bro_object_clone((BroObject *) dst, (BroObject *) src))
    {
      D(("Cloning parent failed.\n"));
      D_RETURN_(FALSE);
    }

  if (src->val_type && ! (dst->val_type = (BroType *) __bro_sobject_copy((BroSObject *) src->val_type)))
    {
      D(("Cloning type failed.\n"));
      D_RETURN_(FALSE);      
    }
  
  if (src->val_attrs && ! (dst->val_attrs = (BroAttrs *) __bro_sobject_copy((BroSObject *) src->val_attrs)))
    {
      D(("Cloning attributes failed.\n"));
      D_RETURN_(FALSE);
    }

  switch (dst->val_type->internal_tag)
    {
    case BRO_INTTYPE_INT:
    case BRO_INTTYPE_UNSIGNED:
    case BRO_INTTYPE_IPADDR:
      /* Hack for ports */
      if (src->val_type->tag == BRO_TYPE_PORT)
	dst->val_port = src->val_port;
      else
	dst->val_int = src->val_int;
      break;
      
    case BRO_INTTYPE_DOUBLE:
      dst->val_double = src->val_double;
      break;
      
    case BRO_INTTYPE_STRING:
      {
	BroString *string = bro_string_copy(&src->val_str);
	dst->val_str = *string;
      }
      break;

    case BRO_INTTYPE_SUBNET:
      dst->val_subnet = src->val_subnet;
      break;

    case BRO_INTTYPE_OTHER:
      /* That's okay, handled in subtype */
      break;
      
    default:
      D(("Unknown internal type tag: %i\n", dst->val_type->internal_tag));
      D_RETURN_(FALSE);
    }  
  
  D_RETURN_(TRUE);
}


void *
__bro_val_get(BroVal *val)
{
  D_ENTER;

  /* Following the comments in broccoli.h, we return atomic values
   * as copies into *result, and complex types (i.e., structs) have
   * all members assigned to point to internal values so the user
   * does not have to clean up the returned value. The user can 
   * still keep those values around if necessary by copying them.
   */
  if (! val->val_type)
    {
      D(("No type in val %p\n", val));
      D_RETURN_(NULL);
    }
  
   switch (val->val_type->tag)    
    {
    case BRO_TYPE_BOOL:
    case BRO_TYPE_INT:
    case BRO_TYPE_ENUM:
    case BRO_TYPE_COUNT:
    case BRO_TYPE_COUNTER:
    case BRO_TYPE_IPADDR:
    case BRO_TYPE_NET:
      D_RETURN_(&val->val_int);
      
    case BRO_TYPE_PORT:
      D_RETURN_(&val->val_port);
      
    case BRO_TYPE_DOUBLE:
    case BRO_TYPE_TIME:
    case BRO_TYPE_INTERVAL:
      D_RETURN_(&val->val_double);
      
    case BRO_TYPE_STRING:
      D_RETURN_(&val->val_str);
      
    case BRO_TYPE_SUBNET:
      D_RETURN_(&val->val_subnet);
      
    case BRO_TYPE_RECORD:
      D(("Inheritance broken -- record types should not be handled here.\n"));
      D_RETURN_(NULL);
      
    default:
      D(("Type %i currently not extractable.\n", val->val_type->tag));
    }
   
   D_RETURN_(NULL);
}


BroMutableVal *
__bro_mutable_val_new(void)
{
  BroMutableVal *val;

  D_ENTER;

  if (! (val = calloc(1, sizeof(BroMutableVal))))
    D_RETURN_(NULL);

  __bro_mutable_val_init(val);

  D_RETURN_(val);
}


void
__bro_mutable_val_init(BroMutableVal *mv)
{
  BroSObject *sobj = (BroSObject *) mv;
  
  D_ENTER;
  
  __bro_val_init((BroVal *) mv);
  
  sobj->read  = (BroSObjectRead) __bro_mutable_val_read;
  sobj->write = (BroSObjectWrite) __bro_mutable_val_write;
  sobj->free  = (BroSObjectFree) __bro_mutable_val_free;
  sobj->clone = (BroSObjectClone) __bro_mutable_val_clone;
  
  sobj->type_id = SER_MUTABLE_VAL;

  /* BroMutableVal inherits __bro_val_get and doesn't override it. */

  D_RETURN;
}


void
__bro_mutable_val_free(BroMutableVal *mv)
{
  D_ENTER;
  
  __bro_sobject_release((BroSObject *) mv->id);
  __bro_val_free((BroVal *) mv);
  
  D_RETURN;
}


int
__bro_mutable_val_read(BroMutableVal *mv, BroConn *bc)
{
  BroString tmp;

  D_ENTER;

  bro_string_init(&tmp);

  if (! __bro_val_read((BroVal *) mv, bc))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_char(bc->rx_buf, &mv->props))
    D_RETURN_(FALSE);

  if (! __bro_buf_read_string(bc->rx_buf, &tmp))
    D_RETURN_(FALSE);

  /* FIXME: now need to obtain real BroID from that name */

  D_RETURN_(TRUE);
}


int
__bro_mutable_val_write(BroMutableVal *mv, BroConn *bc)
{
  D_ENTER;
  
  if (! __bro_val_write((BroVal *) mv, bc))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_char(bc->tx_buf, mv->props))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_string(bc->tx_buf, (mv->id ? &mv->id->name : NULL)))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


int
__bro_mutable_val_clone(BroMutableVal *dst, BroMutableVal *src)
{
  D_ENTER;
  
  if (! __bro_val_clone((BroVal *) dst, (BroVal *) src))
    D_RETURN_(FALSE);
  
  if (src->id && ! (dst->id = (BroID *) __bro_sobject_copy((BroSObject *) src->id)))
    D_RETURN_(FALSE);
  
  src->props = dst->props;
  
  D_RETURN_(TRUE);
}


BroRecordVal *
__bro_record_val_new(void)
{
  BroRecordVal *val;

  D_ENTER;

  if (! (val = calloc(1, sizeof(BroRecordVal))))
    D_RETURN_(NULL);

  __bro_record_val_init(val);

  D_RETURN_(val);
}


void
__bro_record_val_init(BroRecordVal *rv)
{
  BroSObject *sobj = (BroSObject *) rv;
  BroVal *val = (BroVal *) rv;

  D_ENTER;
  
  __bro_mutable_val_init((BroMutableVal *) rv);
  
  sobj->read  = (BroSObjectRead) __bro_record_val_read;
  sobj->write = (BroSObjectWrite) __bro_record_val_write;
  sobj->free  = (BroSObjectFree) __bro_record_val_free;
  sobj->clone = (BroSObjectClone) __bro_record_val_clone;

  sobj->type_id = SER_RECORD_VAL;

  val->get_data = (BroValAccessor) __bro_record_val_get;
  
  D_RETURN;
}


void
__bro_record_val_free(BroRecordVal *rv)
{
  D_ENTER;

  if (! rv)
    D_RETURN;
  
  __bro_record_free(rv->rec);
  __bro_mutable_val_free((BroMutableVal *) rv);

  D_RETURN;
}


int
__bro_record_val_read(BroRecordVal *rv, BroConn *bc)
{
  char opt;
  int i, len;
  BroVal *val;
  
  D_ENTER;
  
  if (! __bro_mutable_val_read((BroMutableVal *) rv, bc))
    D_RETURN_(FALSE);

  /* Clean out old vals, if any */
  __bro_record_free(rv->rec);
  
  if (! (rv->rec = __bro_record_new()))
    D_RETURN_(FALSE);
  
  /* Read in new vals */

  if (! __bro_buf_read_int(bc->rx_buf, &len))
    goto error_return;
  
  for (i = 0; i < len; i++)
    {
      D(("Reading val %i/%i into record %p of val %p\n",
	 i+1, len, rv->rec, rv));

      if (! (__bro_buf_read_char(bc->rx_buf, &opt)))
	goto error_return;
      
      if (opt)
	{
	  if (! (val = (BroVal *) __bro_sobject_unserialize(SER_IS_VAL, bc)))
	    goto error_return;
	}
      else
	{
	  /* We need an empty val if none was given in order to maintain
	   * a chain of vals nonetheless -- the missing type in this new
	   * val indicates that it is an unassigned val.
	   */
	  D(("WARNING -- unassigned val.\n"));
	  if (! (val = __bro_val_new()))
	    goto error_return;
	}
      
      __bro_record_add_val(rv->rec, val);
    }
  
  D_RETURN_(TRUE);

 error_return:
  __bro_record_free(rv->rec);
  rv->rec = NULL;
  D_RETURN_(FALSE);
}


int
__bro_record_val_write(BroRecordVal *rv, BroConn *bc)
{
  BroVal *val;
  int i;

  D_ENTER;

  if (! __bro_mutable_val_write((BroMutableVal *) rv, bc))
    D_RETURN_(FALSE);
  
  if (! __bro_buf_write_int(bc->tx_buf, rv->rec->val_len))
    D_RETURN_(FALSE);
  
  if (! rv->rec && rv->rec->val_len > 0)
    D_RETURN_(FALSE);
  
  D(("Writing out %i vals in record %p.\n", rv->rec->val_len, rv->rec));

  for (i = 0, val = rv->rec->val_list.tqh_first; val; i++, val = val->val_list.tqe_next)
    {
      D(("Val %i/%p's type: %p\n", i, val, val->val_type));

      if (! __bro_buf_write_char(bc->tx_buf, (val->val_type ? 1 :0)))
	D_RETURN_(FALSE);
      
      if (val->val_type)
	{	  
	  if (! __bro_sobject_serialize((BroSObject *) val, bc))
	    D_RETURN_(FALSE);
	}
    }
  
  D_RETURN_(TRUE);
}


int
__bro_record_val_clone(BroRecordVal *dst, BroRecordVal *src)
{
  D_ENTER;
  
  if (! __bro_mutable_val_clone((BroMutableVal *) dst, (BroMutableVal *) src))
    D_RETURN_(FALSE);
  
  if (src->rec && ! (dst->rec = __bro_record_copy(src->rec)))
    D_RETURN_(FALSE);
  
  D_RETURN_(TRUE);
}


void *
__bro_record_val_get(BroRecordVal *rv)
{
  return rv->rec;
}
