/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004-2005 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

#include <broccoli.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

char *host_default = "127.0.0.1";
char *port_default = "47758";
char *host_str;
char *port_str;

int count = -1;
int seq;

static void
usage(void)
{
  printf("broping - sends ping events to a Bro agent, expecting pong events.\n"
	 "USAGE: broping [-h|-?] [-d] [-r] [-c num] [-p port] host\n");
  exit(0);
}


static void
bro_pong(BroConn *conn, double *src_time, double *dst_time, uint32 *seq)
{
  double now = bro_util_current_time();

  printf("pong event from %s: seq=%i, time=%f/%f s\n",
	 host_str, *seq, *dst_time - *src_time,
	 now - *src_time);

  conn = NULL;
}

static void
bro_pong_record(BroConn *conn, BroRecord *rec)
{
  double now = bro_util_current_time();
  double *src_time, *dst_time;
  uint32 *seq;

  if (! bro_record_get_nth_val(rec, 0, BRO_TYPE_COUNT, &seq))
    {
      printf("Error getting sequence count from event.\n");
      return;
    }

  if (! bro_record_get_nth_val(rec, 1, BRO_TYPE_TIME, &src_time))
    {
      printf("Error getting src time from event.\n");
      return;
    }
  
  if (! bro_record_get_nth_val(rec, 2, BRO_TYPE_TIME, &dst_time))
    {
      printf("Error getting dst time from event.\n");
      return;
    }
  
  printf("pong event from %s: seq=%i, time=%f/%f s\n",
	 host_str, *seq, *dst_time - *src_time,
	 now - *src_time);
  
  conn = NULL;
}

int
main(int argc, char **argv)
{
  int opt, port, use_record = 0, debugging = 0;
  BroConn *bc;
  extern char *optarg;
  extern int optind;
  char hostname[512];

  host_str = host_default;
  port_str = port_default;

  bro_debug_calltrace = 0;
  bro_debug_messages  = 0;

  while ( (opt = getopt(argc, argv, "c:p:dh?r")) != -1)
    {
      switch (opt)
	{
	case 'd':
	  debugging++;

	  if (debugging == 1)
	    bro_debug_messages = 1;
	  
	  if (debugging > 1)
	    bro_debug_calltrace = 1;
	  break;

	case 'h':
	case '?':
	  usage();

	case 'c':
	  count = strtol(optarg, NULL, 0);
	  if (errno == ERANGE || count < 1)
	    {
	      printf("Please provide an integer to -c.\n");
	      exit(-1);
	    }
	  break;
	  
	case 'p':
	  port_str = optarg;
	  break;

	case 'r':
	  use_record = 1;
	  break;

	default:
	  usage();
	}
    }

  argc -= optind;
  argv += optind;

  if (argc > 0)
    host_str = argv[0];

  /*
  if (! (host = gethostbyname(host_str)) ||
      ! (host->h_addr_list[0]))
    {
      printf("Could not resolve host %s\n", host_str);
      exit(-1);
    }
  */

  port = strtol(port_str, NULL, 0);
  if (errno == ERANGE)
    {
      printf("Please provide a port number with -p.\n");
      exit(-1);
    }

  snprintf(hostname, 512, "%s:%s", host_str, port_str);
  
  /* Connect to Bro */
  if (! (bc = bro_connect_str(hostname, BRO_CFLAG_RECONNECT | BRO_CFLAG_ALWAYS_QUEUE)))
    {
      printf("Could not connect to Bro at %s:%s.\n", host_str, port_str);
      exit(-1);
    }
  
  /* Request "pong" events, and have bro_pong called when they
   * arrive. The callback mechanism automatically figures out
   * the number of arguments and invokes the callback accordingly.
   * Use record-based callback if -r option was given.
   */
  if (use_record)
    bro_event_registry_add(bc, "pong", (BroEventFunc) bro_pong_record);
  else
    bro_event_registry_add(bc, "pong", (BroEventFunc) bro_pong);
  
  bro_event_registry_request(bc);

  if (! bro_conn_await_handshake(bc, 10))
    {
      printf("Could not complete handshake successfully.\n");
      exit(-1);
    }
  
  /* Enter pinging loop */
  for ( ; ; )
    {
      BroEvent *ev;
      
      bro_conn_process_input(bc);
      
      /* Create empty "ping" event */
      if ( (ev = bro_event_new("ping")))
	{
	  double timestamp = bro_util_current_time();

	  if (use_record)
	    {
	      /* Create a record with the sequence number as first
	       * element of type counter, and the second element the
	       * current time:
	       */
	      BroRecord *rec = bro_record_new();
	      
	      bro_record_add_val(rec, "seq", BRO_TYPE_COUNT, &seq);
	      bro_record_add_val(rec, "src_time", BRO_TYPE_TIME, &timestamp);
	      bro_event_add_val(ev, BRO_TYPE_RECORD, rec);
	      
	      bro_record_free(rec);
	    }
	  else
	    {
	      /* Add a timestamp to it: */
	      bro_event_add_val(ev, BRO_TYPE_TIME, &timestamp);
	      
	      /* Add the sequence counter: */
	      bro_event_add_val(ev, BRO_TYPE_COUNT, &seq);
	    }

	  seq++;

	  /* Ship it -- sends it if possible, queues it otherwise */
	  bro_event_send(bc, ev);
	  bro_event_free(ev);
	}

      if (count > 0 && seq == count)
	break;

#ifdef __MINGW32__
      sleep(1000);
#else
      sleep(1);
#endif
    }

  /* Disconnect from Bro */
  bro_disconnect(bc);

  return 0;
}
